/*
 * resolv.h -- DNS resolver interface for the Linux kernel
 * Copyright (C) 2001  Andrew Main
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef _LINUX_RESOLV_H
#define _LINUX_RESOLV_H 1

/*
 * kernel-internal resolver interface
 *
 * Although DNS data formats and assigned numbers are used, the
 * resolver interface is *not* simply an interface to the DNS protocol.
 * It operates at a higher level.  Consequently there are some DNS record
 * types that cannot be used with this resolver interface, because they
 * have meaning only within the context of the DNS protocol itself,
 * rather than labelling data about subject entities.  These include
 * metadata types (e.g., NS, SIG), forwarding records (e.g., CNAME,
 * DNAME), and transaction-data pseudo-types (e.g., OPT, ANY).
 */

#include <asm/types.h>

/*
 * domain names
 *
 * In all interfaces, domain names are represented in the standard DNS form,
 * with compressed labels (and anything else that doesn't make sense in this
 * context) not permitted.  This means that each normal label consists of a
 * length octet in the range [1, 63] followed by that many octets of case-
 * insensitive (but otherwise binary) label.  A domain name consists of a
 * sequence of labels terminated by the root label, which is represented by a
 * zero octet.
 */

typedef u8 *domainname_ptr_t;

#define BASICLABEL_MAXLEN 63

/*
 * The data attached to a domain name is also divided up by class and type.
 */

typedef u16 domainclass_t;
typedef u16 rrtype_t;

/*
 * When an RRset is passed around, it is represented in the following format.
 * There are explicit length fields because the data formats can be variable-
 * length.  Users of this structure are expected to know which domain, class
 * and type it pertains to.
 *
 * The rdata structure may be unaligned.
 */

struct rdata {
	u16 length;   /* length of the data that follows */
	u8 data[0];   /* struct hack */
};

struct rrset {
	u32 size;   /* total size of RRset structure, including this */
	u16 rrcount;   /* number of resource records that follow */
	struct rdata rdata[0];   /* struct hack */
};

/*
 * resolv_getrrset(): look up an RRset and return a copy of the data
 *
 * All relevant forwarding (through CNAMEs and DNAMEs) is performed
 * invisibly; the RRset returned is that for the domain ultimately
 * pointed to.  It follows that a query for forwarding records (such as
 * CNAME) will never return any data -- on a successful lookup an empty
 * RRset would be returned, because the domain ultimately pointed to
 * has no forwarding records.
 *
 * On success, returns a kmalloced rrset structure.  It is the caller's
 * responsibility to free the memory.  It is possible for the RRset to
 * be empty (size = 6, rrcount = 0) -- this indicates that the specified
 * domain does exist, but has no records of the specified type.
 *
 * On error, returns an ERR_PTR indicating the error.  The error number
 * is -ENOENT if the specified domain doesn't exist in the specified
 * class: this is an authoritative statement that it doesn't exist,
 * not just a failure to find data.  Other error numbers indicate a
 * temporary failure somewhere in lookup.
 */

extern struct rrset *resolv_getrrset(domainname_ptr_t domainname,
	domainclass_t domainclass, rrtype_t rrtype);

/*
 * resolv_userrset(): look up an RRset without copying it
 *
 * This does the same lookup as resolv_getrrset(), but doesn't allocate
 * memory to return the data.  Instead, on success a callback function is
 * called, passing in a pointer to the rrset structure.  This structure
 * must not be modified by the callback, and it is valid only during
 * the callback; the idea is that the callback function can extract
 * the information the caller needs, thus avoiding making a copy of the
 * entire RRset.  The return value of resolv_userrset() is whatever the
 * callback function returns.
 *
 * On error, the callback function is not called, and an error number is
 * returned instead.  Error numbers are as they are for resolv_getrrset().
 */

extern int resolv_userrset(domainname_ptr_t domainname,
	domainclass_t domainclass, rrtype_t rrtype,
	int (*cb)(void *cb_magic, struct rrset const *rrset), void *cb_magic);

#endif /* !_LINUX_RESOLV_H */
