/*
 * resolv_dn_util.c -- domain name utilities for the Linux kernel
 * Copyright (C) 2001  Andrew Main
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <linux/module.h>
#include <linux/string.h>
#include <asm/types.h>
#include "resolv_dn_util.h"

ssize_t domainname_text_to_bin(u8 *buf, size_t buflen,
	char const *text, ssize_t textlen)
{
	char const *textend = textlen<0 ? strchr(text, 0) : text+textlen;
	size_t olen = 0;
	if(text == textend)
		return -1;
	if(text+1 == textend && *text == '.')
		text++;
	while(text != textend) {
		size_t ooffs = olen++;
		char c = *text++;
		if(c == '.')
			return -1;
		while(1) {
			if(c == '\\') {
				if(text == textend)
					return -1;
				c = *text++;
				if(c >= '0' && c <= '9') {
					int n;
					if(textend-text < 2)
						return -1;
					n = (c-'0') * 100;
					c = *text++;
					if(c < '0' || c > '9')
						return -1;
					n += (c-'0')*10;
					c = *text++;
					if(c < '0' || c > '9')
						return -1;
					n += c-'0';
					if(n > 255)
						return -1;
					c = (char)n;
				}
			}
			if(olen < buflen)
				buf[olen] = (u8)c;
			olen++;
			if(text == textend)
				break;
			c = *text++;
			if(c == '.')
				break;
			if(olen-ooffs == 64)
				return -1;
		}
		if(olen < buflen)
			buf[ooffs] = olen-ooffs - 1;
	}
	if(olen < buflen)
		buf[olen] = 0;
	return olen+1;
}

size_t domainname_bin_to_text(char *buf, size_t buflen,
	u8 const *bin, char const *alsoescape, int finaldotp, int terminatep)
{
	size_t olen = 0;
	if(!*bin)
		goto finaldot;
	while(1) {
		u8 labellen = *bin++;
		while(labellen--) {
			u8 c = *bin++;
			if(c == '\\' || c == '.') {
				if(olen+2 <= buflen) {
					buf[olen] = '\\';
					buf[olen+1] = c;
				}
				olen += 2;
			} else if(c < '!' || c > '~' || strchr(alsoescape, c)) {
				if(olen+4 <= buflen) {
					buf[olen] = '\\';
					buf[olen+1] = '0' + c/100;
					buf[olen+2] = '0' + (c/10)%10;
					buf[olen+3] = '0' + c%10;
				}
				olen += 4;
			} else {
				if(olen < buflen)
					buf[olen] = c;
				olen++;
			}
		}
		if(!*bin)
			break;
		if(olen < buflen)
			buf[olen] = '.';
		olen++;
	}
	if(finaldotp) {
finaldot:
		if(olen < buflen)
			buf[olen] = '.';
		olen++;
	}
	if(terminatep && olen < buflen)
		buf[olen] = 0;
	return olen;
}

ssize_t domainname_bin_check_length(u8 const *bin, size_t maxlen)
{
	size_t len = 0;
	while(1) {
		u8 o;
		if(len >= maxlen)
			return -1;
		o = bin[len++];
		if(!o)
			return len;
		else if(o < 64)
			len += o;
		else
			return -1;
	}
}

/*
 * module stuff
 */

EXPORT_SYMBOL(domainname_text_to_bin);
EXPORT_SYMBOL(domainname_bin_to_text);
EXPORT_SYMBOL(domainname_bin_check_length);

#ifdef MODULE_AUTHOR
MODULE_AUTHOR("Andrew Main (Zefram) <zefram@fysh.org>");
#endif /* MODULE_AUTHOR */
#ifdef MODULE_LICENSE
MODULE_LICENSE("GPL");
#endif /* MODULE_LICENSE */
#ifdef MODULE_DESCRIPTION
MODULE_DESCRIPTION("Domain name utility functions");
#endif /* MODULE_DESCRIPTION */
