/*
 * resolv_dn_util.h -- domain name utilities for the Linux kernel
 * Copyright (C) 2001  Andrew Main
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef _LINUX_RESOLV_DN_UTIL_H
#define _LINUX_RESOLV_DN_UTIL_H 1

/*
 * domain name formats
 *
 * The resolver works with domain names only in the standard binary format
 * defined by DNS (RFC 1035): it uses the binary form both in its function
 * interfaces and also embedded in resource data.  The functions here
 * are for code that needs to manipulate domain names in binary format
 * but does not need to know any details of that format.
 *
 * These functions deal only with the canonical binary domain name format,
 * not the `compressed' formats that are used in DNS packets.
 *
 * The things you need to know about domain names:
 *
 * * the binary format consists of a sequence of one or more consecutive
 *   octets; any octet value can occur, including NUL
 *
 * * there is significant internal syntax in the binary form, so it is
 *   possible for it to be malformed
 *
 * * reading from the start of a binary domain name, it is possible to
 *   determine its length; it cannot be read from the end
 *
 * * there is a standard textual format for domain names, consisting of
 *   a string of one or more characters; for input it can include any
 *   characters whatsoever, including NUL, but on output it is usually
 *   restricted to ASCII graphical characters, and it is possible to avoid
 *   using any particular characters, other than ".", "\", and digits
 *
 * * the textual format has significant internal syntax, and so can
 *   be malformed
 *
 * * the binary form of a domain name is never longer than the length
 *   of the textual format + 2 bytes; the textual form of a domain name
 *   can be much longer than the binary
 */

#include <asm/types.h>

/*
 * domainname_text_to_bin(): interpret a textual domain name
 *
 * The input is pointed to by `text', and is either of length `textlen'
 * and may contain NUL bytes, or if textlen==-1 then it is NUL-terminated.
 * It is a domain name in textual master file format; all standard
 * \-escapes are permitted, and a trailing dot is optional, except that
 * the root domain must be named by exactly one dot.
 *
 * The output buffer is pointed to by `buf', and is of length `buflen'.
 * If there is enough space, the binary form of the domain name is written
 * into the start of the buffer, the rest of the buffer is untouched,
 * and the length of the output is returned.  If there is insufficient
 * space, the contents of the buffer is unspecified, and the length of
 * the output that would have been produced (>buflen) is returned.
 *
 * If there is a syntax error in the input, the contents of the buffer
 * is unspecified, and -1 is returned.  This occurs regardless of
 * buffer size.
 */

ssize_t domainname_text_to_bin(u8 *buf, size_t buflen,
	char const *text, ssize_t textlen);

/*
 * domainname_bin_to_text(): render a domain name in textual form
 *
 * The input is a binary-format domain name pointed to by `bin'.  It must
 * be well-formed.
 *
 * This domain name is rendered into textual master file format,
 * using only ASCII graphical characters, and avoiding the use of any
 * characters that appear in the string `alsoescape'; it is not possible
 * to avoid the use of the characters ".", "\" and digits if the domain
 * name requires them.  A final dot is appended to the textual form if
 * and only if `finaldotp' is non-zero; however, the root domain name is
 * always represented as exactly one dot.  A NUL terminator is appended
 * to the output string if and only if `terminatep' is non-zero.
 *
 * The output buffer is pointed to by `buf', and is of length `buflen'.
 * If there is enough space, the textual form of the domain name is
 * written into the start of the buffer and the rest of the buffer is
 * untouched.  If there is insufficient space, the contents of the buffer
 * is unspecified.  In either case, the length of the proper output,
 * excluding the NUL terminator, is returned.  It follows that the
 * test of whether there was enough space in the buffer is "if(result +
 * !!terminatep <= buflen)".
 */

size_t domainname_bin_to_text(char *buf, size_t buflen,
	u8 const *bin, char const *alsoescape, int finaldotp, int terminatep);

/*
 * domainname_bin_check_length(): check syntax of a binary-format domain name
 *
 * This function checks the well-formedness of a binary-format domain
 * name.  The input is pointed to by `bin', and is supposedly a domain
 * name in binary format, which may be up to `maxlen' in length.  If the
 * domain name is well-formed and does not exceed that length, then the
 * true length is returned.  Otherwise, -1 is returned.
 *
 * A typical usage is that one has a message that includes a binary-format
 * domain name followed by other data, with the message format relying
 * on the reader being able to determine the extent of the domain name
 * in order to locate the following data.  In this case `bin' would
 * point to the known start of the domain name, and `maxlen' would be
 * the distance to the known end of the message (or the end of the area
 * where the domain name could extend to).
 */

ssize_t domainname_bin_check_length(u8 const *bin, size_t maxlen);

#endif /* !_LINUX_RESOLV_UTIL_H */
