/*
 * resolvfs.c -- domain name resolution virtual filesystem for Linux
 * Copyright (C) 2001  Andrew Main
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * domain name resolution virtual filesystem
 *
 * Valid pathnames in thise filesystem are of the form
 *   /<class>/<domain>/<rrtype>
 * where <class> and <rrtype> are case-insensitive class/RRtype names.  The
 * contents of the file thus named is the conventional "master file" syntax
 * representation of the resource records referred to.
 */

/*
 * TODO: privacy mode: randomised inode numbers and no cached dentries,
 * to prevent analysis of other users' usage patterns.  note for
 * documentation: this reduces the privacy problem to whatever exists
 * with the underlying DNS lookup system; covert channels are inherent
 * to any shared caching system.
 */

#include <linux/module.h>
#include <linux/init.h>
#include <linux/fs.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/stddef.h>
#include <asm/uaccess.h>
#include <asm/unaligned.h>
#include <asm/byteorder.h>
#include "resolv.h"
#include "resolv_numbers.h"
#include "resolv_dn_util.h"

#define ARRAY_LEN(A) (sizeof(A) / sizeof(*(A)))

/*
 * number interpretation
 *
 * scan_u16() interprets an unsigned decimal number, requiring it to be
 * within the u16 range.  Returns 0 on success, or -1 if the number has
 * bad format or exceeds the permitted range.
 */

static int scan_u16(char const *str, u16 *result_p)
{
	u32 result = 0;
	char c = *str;
	do {
		if(c < '0' || c > '9')
			return -1;
		result = result*10 + c-'0';
		if(result > 0xffff)
			return -1;
	} while((c = *++str));
	*result_p = result;
	return 0;
}

/*
 * text formatting of parts of RRs
 *
 * These functions write into the provided output buffer, which starts at
 * `buf' and is of length `buflen'.  They never write outside that buffer.
 * They start writing at the offset `offset'; if offset > buflen then
 * nothing is written.  They are guaranteed to produce correct results
 * in the output buffer only if there is enough space for what they
 * want to write.  Otherwise, the contents of the buffer past `offset'
 * are undefined.
 *
 * The functions return the offset within the buffer just past the end of
 * their output; this is offset + (length of output).  This is calculated
 * correctly even if there was insufficient space in the buffer for the
 * output; in this case, the offset returned will point outside the buffer
 * (will be >buflen).
 */

static size_t render_char(char *buf, size_t buflen, size_t offset, char c)
{
	if(offset < buflen)
		buf[offset] = c;
	return offset + 1;
}

static size_t render_block(char *buf, size_t buflen, size_t offset,
	char const *data, size_t datalen)
{
	if(offset + datalen <= buflen)
		memcpy(buf+offset, data, datalen);
	return offset + datalen;
}

static size_t render_unsigned(char *buf, size_t buflen, size_t offset,
	u32 number)
{
	char text[11];
	size_t tlen = sprintf(text, "%u", number);
	return render_block(buf, buflen, offset, text, tlen);
}

static size_t render_ipv4addr(char *buf, size_t buflen, size_t offset,
	u8 const *data)
{
	char text[12];
	size_t tlen;
	tlen = sprintf(text, "%d.%d.%d.%d", data[0], data[1], data[2], data[3]);
	return render_block(buf, buflen, offset, text, tlen);
}

static size_t render_ipv6addr(char *buf, size_t buflen, size_t offset,
	u8 const *data)
{
	char text[40];
	u16 words[8];
	int embedv4, ew, i;
	char *tptr = text;
	int bestbase = -1, bestend = -1, bestlen = 0, thisbase = 0;
	for(i = 0; i != 8; i++)
		words[i] = ntohs(get_unaligned((u16 *)(data + 2*i)));
	embedv4 = !words[0] && !words[1] && !words[2] && !words[3] &&
		!words[4] && (words[5] == 0xffff ||
				(!words[5] && (words[6] || words[7] > 1)));
	ew = embedv4 ? 6 : 8;
	while(thisbase != ew) {
		if(words[thisbase]) {
			thisbase++;
			continue;
		}
		for(i = thisbase+1; i!=ew && !words[i]; i++) ;
		if(i-thisbase > bestlen ||
			(i-thisbase == bestlen &&
				bestbase==0 && i!=8)) {
			bestbase = thisbase;
			bestend = i;
			bestlen = i-thisbase;
		}
		thisbase = i;
	}
	if(bestbase == 0)
		*tptr++ = ':';
	for(i = 0; ; ) {
		if(i == bestbase) {
			*tptr++ = ':';
			i = bestend;
			if(i == 8)
				break;
		}
		if(i == 6 && embedv4) {
			tptr += sprintf(tptr, "%d.%d.%d.%d",
					words[6] >> 8, words[6] & 0xff,
					words[7] >> 8, words[7] & 0xff);
			break;
		}
		tptr += sprintf(tptr, "%x", (unsigned)words[i]);
		if(++i == 8)
			break;
		*tptr++ = ':';
	}
	return render_block(buf, buflen, offset, text, tptr-text);
}

static size_t render_textstr(char *buf, size_t buflen, size_t offset,
	u8 const *data)
{
	size_t ilen = *data++;
	offset = render_char(buf, buflen, offset, '"');
	while(ilen--) {
		u8 c = *data++;
		if(c == '\\' || c == '"') {
			offset = render_char(buf, buflen, offset, '\\');
			offset = render_char(buf, buflen, offset, (char)c);
		} else if(c >= ' ' || c <= '~') {
			offset = render_char(buf, buflen, offset, (char)c);
		} else {
			offset = render_char(buf, buflen, offset, '\\');
			offset = render_char(buf, buflen, offset, '0'+c/100);
			offset = render_char(buf, buflen, offset, '0'+c/10%10);
			offset = render_char(buf, buflen, offset, '0'+c%10);
		}
	}
	return render_char(buf, buflen, offset, '"');
}

static size_t render_hex(char *buf, size_t buflen, size_t offset,
	u8 const *data, size_t datalen)
{
	if(offset + 2*datalen <= buflen) {
		size_t dl;
		buf += offset;
		for(dl = datalen; dl--; data++) {
			static char const xdigit[16] = "0123456789abcdef";
			*buf++ = xdigit[*data >> 4];
			*buf++ = xdigit[*data & 0xf];
		}
	}
	return offset + 2*datalen;
}

static size_t render_domainname(char *buf, size_t buflen, size_t offset,
	u8 const *dn)
{
	if(offset <= buflen) {
		buf += offset;
		buflen -= offset;
	} else {
		buf = NULL;
		buflen = 0;
	}
	return offset + domainname_bin_to_text(buf, buflen, dn, "", 1, 0);
}

/*
 * text formatting of resource records
 *
 * rr_render_fn is a function interface used for rendering resource
 * records.  The resource data passed in is rendered in a textual format
 * in the buffer provided, without any termination, and the length of the
 * rendered text is returned.  If there is insufficient buffer space,
 * then the contents of the buffer is undefined, and the buffer size
 * required is returned.
 */

typedef size_t (*rr_render_fn)(char *buf, size_t buflen,
					struct rdata const *rdata);

static size_t render_generic(char *buf, size_t buflen,
	struct rdata const *rdata)
{
	unsigned dlen = get_unaligned(&rdata->length);
	char start[10];
	int startlen;
	size_t offset;
	if(dlen == 0)
		return render_block(buf, buflen, 0, "\\# 0", 4);
	startlen = sprintf(start, "\\# %u ", dlen);
	offset = render_block(buf, buflen, 0, start, startlen);
	return render_hex(buf, buflen, offset, rdata->data, dlen);
}

static size_t render_nsap(char *buf, size_t buflen, struct rdata const *rdata)
{
	unsigned dlen = get_unaligned(&rdata->length);
	if(!dlen)
		return render_generic(buf, buflen, rdata);
	return render_hex(buf, buflen, 0, rdata->data, dlen);
}

static size_t render_a(char *buf, size_t buflen, struct rdata const *rdata)
{
	if(get_unaligned(&rdata->length) != 4)
		return render_generic(buf, buflen, rdata);
	return render_ipv4addr(buf, buflen, 0, rdata->data);
}

static size_t render_aaaa(char *buf, size_t buflen, struct rdata const *rdata)
{
	if(get_unaligned(&rdata->length) != 16)
		return render_generic(buf, buflen, rdata);
	return render_ipv6addr(buf, buflen, 0, rdata->data);
}

static size_t render_strings(char *buf, size_t buflen,
	struct rdata const *rdata, size_t minstrings, size_t maxstrings)
{
	size_t datalen = get_unaligned(&rdata->length), di, offset = 0;
	u8 *data = rdata->data;
	size_t nstrings;
	for(nstrings = 0, di = 0; di != datalen; nstrings++) {
		di += 1 + data[di];
		if(di > datalen)
			goto badfmt;
	}
	if(nstrings < minstrings || nstrings > maxstrings) {
badfmt:
		return render_generic(buf, buflen, rdata);
	}
	while(1) {
		size_t len;
		offset = render_textstr(buf, buflen, offset, data);
		len = 1 + *data;
		data += len;
		datalen -= len;
		if(!datalen)
			return offset;
		offset = render_char(buf, buflen, offset, ' ');
	}
}

static size_t render_txt(char *buf, size_t buflen, struct rdata const *rdata)
{
	return render_strings(buf, buflen, rdata, 1, (size_t)-1);
}

static size_t render_x25(char *buf, size_t buflen, struct rdata const *rdata)
{
	return render_strings(buf, buflen, rdata, 1, 1);
}

static size_t render_hinfo(char *buf, size_t buflen, struct rdata const *rdata)
{
	return render_strings(buf, buflen, rdata, 2, 2);
}

static size_t render_isdn(char *buf, size_t buflen, struct rdata const *rdata)
{
	return render_strings(buf, buflen, rdata, 1, 2);
}

static size_t render_gpos(char *buf, size_t buflen, struct rdata const *rdata)
{
	return render_strings(buf, buflen, rdata, 3, 3);
}

static size_t render_wks(char *buf, size_t buflen, struct rdata const *rdata)
{
	u8 *data = rdata->data;
	size_t datalen = get_unaligned(&rdata->length);
	size_t offset;
	u16 portno;
	if(datalen < 5)
		return render_generic(buf, buflen, rdata);
	offset = render_ipv4addr(buf, buflen, 0, data);
	offset = render_char(buf, buflen, offset, ' ');
	offset = render_unsigned(buf, buflen, offset, data[4]);
	for(portno = 0; portno != ((datalen-5)<<3); portno++) {
		if(data[5+(portno>>3)] & (0x80 >> (portno&7))) {
			offset = render_char(buf, buflen, offset, ' ');
			offset = render_unsigned(buf, buflen, offset, portno);
		}
	}
	return offset;
}

static size_t render_ptr(char *buf, size_t buflen, struct rdata const *rdata)
{
	u8 *data = rdata->data;
	size_t datalen = get_unaligned(&rdata->length);
	if(datalen != domainname_bin_check_length(data, datalen))
		return render_generic(buf, buflen, rdata);
	return render_domainname(buf, buflen, 0, data);
}

static size_t render_mx(char *buf, size_t buflen, struct rdata const *rdata)
{
	size_t offset;
	u8 *data = rdata->data;
	size_t datalen = get_unaligned(&rdata->length);
	if(datalen < 2 || datalen-2 != domainname_bin_check_length(data+2,
								datalen-2))
		return render_generic(buf, buflen, rdata);
	offset = render_unsigned(buf, buflen, 0,
					ntohs(get_unaligned((u16 *)data)));
	offset = render_char(buf, buflen, offset, ' ');
	return render_domainname(buf, buflen, offset, data+2);
}

static size_t render_px(char *buf, size_t buflen, struct rdata const *rdata)
{
	size_t offset;
	u8 *data = rdata->data;
	size_t datalen = get_unaligned(&rdata->length);
	ssize_t len1;
	if(datalen < 2 ||
		(len1 = domainname_bin_check_length(data+2, datalen-2)) == -1 ||
		datalen != 2 + len1 + domainname_bin_check_length(data+2+len1,
								datalen-2-len1))
		return render_generic(buf, buflen, rdata);
	offset = render_unsigned(buf, buflen, 0,
					ntohs(get_unaligned((u16 *)data)));
	offset = render_char(buf, buflen, offset, ' ');
	offset = render_domainname(buf, buflen, offset, data+2);
	offset = render_char(buf, buflen, offset, ' ');
	return render_domainname(buf, buflen, offset, data+2+len1);
}

static size_t render_srv(char *buf, size_t buflen, struct rdata const *rdata)
{
	int n;
	size_t offset = 0;
	u8 *data = rdata->data;
	size_t datalen = get_unaligned(&rdata->length);
	if(datalen < 6 || datalen-6 != domainname_bin_check_length(data+6,
								datalen-6))
		return render_generic(buf, buflen, rdata);
	for(n = 0; n != 3; n++) {
		offset = render_unsigned(buf, buflen, offset,
				ntohs(get_unaligned((u16 *)(data+2*n))));
		offset = render_char(buf, buflen, offset, ' ');
	}
	return render_domainname(buf, buflen, offset, data+6);
}

static size_t render_minfo(char *buf, size_t buflen, struct rdata const *rdata)
{
	size_t offset;
	u8 *data = rdata->data;
	size_t datalen = get_unaligned(&rdata->length);
	ssize_t len1 = domainname_bin_check_length(data, datalen);
	if(len1 == -1 ||
		datalen != len1 +
			domainname_bin_check_length(data+len1, datalen-len1))
		return render_generic(buf, buflen, rdata);
	offset = render_domainname(buf, buflen, 0, data);
	offset = render_char(buf, buflen, offset, ' ');
	return render_domainname(buf, buflen, offset, data+len1);
}

/* TODO: more specific render functions */

/*
 * information about classes and RRtypes
 *
 * struct rrtype_ops represents information about how to handle a
 * particular type of resource records (how the RRdata is formatted).
 * struct rrtype_name represents information about an RRtype name:
 * which RRtype number is represented, and how to handle that RRtype
 * (the rrtype_ops).
 *
 * The name table has names in ASCII sorted order, to facilitate binary
 * search lookups.
 */

struct rrtype_ops {
	rr_render_fn render;
};

#define MAX_RRTYPE_NAME_LEN 8
struct rrtype_name {
	char name[MAX_RRTYPE_NAME_LEN+1];
	rrtype_t rrtype_number;
	struct rrtype_ops rrtype_ops;
};

static struct rrtype_name const rrtype_names[] = {
	{ "a",		T_A,		{render_a} },
	{ "aaaa",	T_AAAA,		{render_aaaa} },
	{ "afsdb",	T_AFSDB,	{render_mx} },
	{ "apl",	T_APL,		{render_generic/*TODO*/} },
	{ "atma",	T_ATMA,		{render_generic/*TODO*/} },
	{ "cert",	T_CERT,		{render_generic/*TODO*/} },
	{ "eid",	T_EID,		{render_generic} },
	{ "gid",	T_GID,		{render_generic} },
	{ "gpos",	T_GPOS,		{render_gpos} },
	{ "hinfo",	T_HINFO,	{render_hinfo} },
	{ "isdn",	T_ISDN,		{render_isdn} },
	{ "kx",		T_KX,		{render_mx} },
	{ "loc",	T_LOC,		{render_generic/*TODO*/} },
	{ "mb",		T_MB,		{render_ptr} },
	{ "md",		T_MD,		{render_ptr} },
	{ "mf",		T_MF,		{render_ptr} },
	{ "mg",		T_MG,		{render_ptr} },
	{ "minfo",	T_MINFO,	{render_minfo} },
	{ "mr",		T_MR,		{render_ptr} },
	{ "mx",		T_MX,		{render_mx} },
	{ "naptr",	T_NAPTR,	{render_generic/*TODO*/} },
	{ "nimloc",	T_NIMLOC,	{render_generic} },
	{ "nsap",	T_NSAP,		{render_nsap} },
	{ "nsap-ptr",	T_NSAP_PTR,	{render_generic} },
	{ "null",	T_NULL,		{render_generic} },
	{ "ptr",	T_PTR,		{render_ptr} },
	{ "px",		T_PX,		{render_px} },
	{ "rp",		T_RP,		{render_minfo} },
	{ "rt",		T_RT,		{render_mx} },
	{ "sink",	T_SINK,		{render_generic} },
	{ "srv",	T_SRV,		{render_srv} },
	{ "txt",	T_TXT,		{render_txt} },
	{ "uid",	T_UID,		{render_generic} },
	{ "uinfo",	T_UINFO,	{render_generic} },
	{ "unspec",	T_UNSPEC,	{render_generic} },
	{ "wks",	T_WKS,		{render_wks} },
	{ "x25",	T_X25,		{render_x25} },
};

/*
 * information about classes
 *
 * The name table has names in ASCII sorted order, to facilitate binary
 * search lookups.
 */

#define MAX_CLASS_NAME_LEN 2
struct class_name {
	char name[MAX_CLASS_NAME_LEN+1];
	domainclass_t class_number;
};

static struct class_name const class_names[] = {
	{ "ch", C_CH },
	{ "hs", C_HS },
	{ "in", C_IN },
};

/*
 * case-insensitive name comparison
 *
 * Strings are compared with Latin majuscules folded to lowercase.  Other
 * characters are not modified.
 */

static int namencmp(char const *a, char const *b, size_t len)
{
	while(len--) {
		int aa = (unsigned char)*a++;
		int bb = (unsigned char)*b++;
		if(!aa && !bb)
			return 0;
		if(aa >= 'A' && aa <= 'Z')
			aa += 'a'-'A';
		if(bb >= 'A' && bb <= 'Z')
			bb += 'a'-'A';
		if(aa != bb)
			return aa-bb;
	}
	return 0;
}

static int namecmp(char const *a, char const *b)
{
	while(1) {
		int aa = (unsigned char)*a++;
		int bb = (unsigned char)*b++;
		if(!aa && !bb)
			return 0;
		if(aa >= 'A' && aa <= 'Z')
			aa += 'a'-'A';
		if(bb >= 'A' && bb <= 'Z')
			bb += 'a'-'A';
		if(aa != bb)
			return aa-bb;
	}
}

/* per-filesystem data */

#define MY_SUPER_DATA(S) ((struct my_super_data *)((S)->u.generic_sbp))
struct my_super_data {
	spinlock_t nextinode_lock;
	ino_t nextinode;
};

static inline ino_t allocate_inum(struct super_block *s)
{
	/* TODO: handle inode number wraparound? */
	ino_t inum;
	spin_lock(&MY_SUPER_DATA(s)->nextinode_lock);
	inum = MY_SUPER_DATA(s)->nextinode++;
	spin_unlock(&MY_SUPER_DATA(s)->nextinode_lock);
	return inum;
}

/*
 * per-inode data
 *
 * This data is different at each level in the filesystem hierarchy.
 * Each level chains to its parent, so that all the relevant data is
 * accessible.
 */

#define CLASS_INODE_DATA(I) ((struct class_inode_data *)((I)->u.generic_ip))
struct class_inode_data {
	domainclass_t class_number;
};

#define DOMAIN_INODE_DATA(I) ((struct domain_inode_data *)((I)->u.generic_ip))
struct domain_inode_data {
	struct class_inode_data *class;
	u8 domain_name[0];   /* struct hack */
};

#define RRSET_INODE_DATA(I) ((struct rrset_inode_data *)((I)->u.generic_ip))
struct rrset_inode_data {
	struct domain_inode_data *domain;
	rrtype_t rrtype_number;
	struct rrtype_ops const *rrtype_ops;
};

/* RRset files */

#define RRSET_FILE_DATA(F) ((struct rrset_file_data *)((F)->private_data))
struct rrset_file_data {
	size_t mfdata_length;
	char *mfdata;
};

struct resolve_rrset {
	struct rrset_file_data *rfd;
	rr_render_fn render;
};

static int resolve_rrset_cb(void *v_magic, struct rrset const *rrset)
{
	struct resolve_rrset *magic = v_magic;
	size_t buflen = 0;
	int rrcount = rrset->rrcount;
	struct rdata const *rdp;
	char *buf;
	if(!rrcount) {
		magic->rfd->mfdata = NULL;
		magic->rfd->mfdata_length = 0;
		return 0;
	}
	for(rdp = rrset->rdata; rrcount--; ) {
		buflen += magic->render(NULL,  0, rdp) + 1;
		rdp = (struct rdata *)(rdp->data + get_unaligned(&rdp->length));
	}
	buf = kmalloc(buflen, GFP_KERNEL);
	if(!buf)
		return -ENOMEM;
	magic->rfd->mfdata = buf;
	magic->rfd->mfdata_length = buflen;
	for(rrcount = rrset->rrcount, rdp = rrset->rdata; rrcount--; ) {
		buf += magic->render(buf,  buflen, rdp);
		*buf++ = '\n';
		rdp = (struct rdata *)(rdp->data + get_unaligned(&rdp->length));
	}
	return 0;
}

static int resolve_rrset(struct rrset_file_data *rfd,
	struct rrset_inode_data *rd)
{
	struct resolve_rrset magic;
	struct domain_inode_data *dd = rd->domain;
	struct class_inode_data *cd = dd->class;
	magic.rfd = rfd;
	magic.render = rd->rrtype_ops->render;
	return resolv_userrset(dd->domain_name, cd->class_number,
			rd->rrtype_number, resolve_rrset_cb, &magic);
}

static int rrset_open(struct inode *i, struct file *f)
{
	int result;
	struct rrset_file_data *rfd = kmalloc(sizeof(*rfd), GFP_KERNEL);
	if(!rfd)
		return -ENOMEM;
	result = resolve_rrset(rfd, RRSET_INODE_DATA(i));
	if(result)
		goto err_read_failed;
	RRSET_FILE_DATA(f) = rfd;
	return 0;
err_read_failed:
	kfree(rfd);
	return result;
}

static int rrset_release(struct inode *i, struct file *f)
{
	struct rrset_file_data *rfd = RRSET_FILE_DATA(f);
	kfree(rfd->mfdata);
	kfree(rfd);
	return 0;
}

static ssize_t rrset_read(struct file *f, char *buf, size_t count, loff_t *ppos)
{
	struct rrset_file_data *rfd = RRSET_FILE_DATA(f);
	if(*ppos >= rfd->mfdata_length)
		return 0;
	if(count > rfd->mfdata_length - *ppos)
		count = rfd->mfdata_length - *ppos;
	copy_to_user(buf, rfd->mfdata + *ppos, count);
	*ppos += count;
	return count;
}

static struct inode_operations rrset_inode_ops = {
};

static struct file_operations rrset_file_ops = {
	open: rrset_open,
	release: rrset_release,
	read: rrset_read,
};

static inline struct inode *mk_rrset_inode(struct super_block *s,
	struct domain_inode_data *domain,
	rrtype_t rrtype_number, struct rrtype_ops const *rrtype_ops)
{
	struct rrset_inode_data *rd;
	struct inode *i;
	rd = kmalloc(sizeof(*rd), GFP_KERNEL);
	if(!rd)
		goto fail_nodata;
	rd->domain = domain;
	rd->rrtype_number = rrtype_number;
	rd->rrtype_ops = rrtype_ops;
	i = iget(s, allocate_inum(s));
	if(!i)
		goto fail_noinode;
	i->i_op = &rrset_inode_ops;
	i->i_fop = &rrset_file_ops;
	i->i_mode = S_IFREG|0444;
	RRSET_INODE_DATA(i) = rd;
	return i;
fail_noinode:
	kfree(rd);
fail_nodata:
	return ERR_PTR(-ENOMEM);
}

/* domain directories */

static struct dentry *domain_lookup(struct inode *parenti,
	struct dentry *childd)
{
	int a, b;
	if(!namencmp(childd->d_name.name, "type", 4)) {
		struct inode *i;
		u16 n;
		static struct rrtype_ops const ro = { render_generic };
		if(scan_u16(childd->d_name.name + 4, &n))
			goto noent;
		i = mk_rrset_inode(parenti->i_sb, DOMAIN_INODE_DATA(parenti),
				n, &ro);
		if(IS_ERR(i))
			return ERR_PTR(PTR_ERR(i));
		d_add(childd, i);
		return NULL;
	}
	for(a = 0, b = ARRAY_LEN(rrtype_names); a != b; ) {
		int m = (a+b)/2;
		int cmp = namecmp(childd->d_name.name, rrtype_names[m].name);
		if(!cmp) {
			/* TODO: do case folding more neatly -- get right inode number */
			struct inode *i = mk_rrset_inode(parenti->i_sb,
						DOMAIN_INODE_DATA(parenti),
						rrtype_names[m].rrtype_number,
						&rrtype_names[m].rrtype_ops);
			if(IS_ERR(i))
				return ERR_PTR(PTR_ERR(i));
			d_add(childd, i);
			return NULL;
		}
		if(cmp < 0)
			b = m;
		else
			a = m+1;
	}
noent:
	d_add(childd, NULL);
	return NULL;
}

static int domain_readdir(struct file *filp, void *dirent, filldir_t filldir)
{
	loff_t pos;
	for(pos = filp->f_pos; pos < 2+ARRAY_LEN(rrtype_names); pos++) {
		if(pos == 0) {
			if(filldir(dirent, ".", 1, 0,
					filp->f_dentry->d_inode->i_ino, DT_DIR))
				break;
		} else if(pos == 1) {
			if(filldir(dirent, "..", 2, 1,
					filp->f_dentry->d_parent->
						d_inode->i_ino, DT_DIR))
				break;
		} else {
			ino_t inum;
			struct rrtype_name const *rnp = &rrtype_names[pos-2];
			char const *name = rnp->name;
			int namelen = strlen(name);
			struct dentry *de =
				lookup_one_len(name, filp->f_dentry, namelen);
			if(IS_ERR(de))
				return PTR_ERR(de);
			inum = de->d_inode->i_ino;
			dput(de);
			if(filldir(dirent, name, namelen, pos, inum, DT_REG))
				break;
		}
	}
	filp->f_pos = pos;
	return 0;
}

static struct inode_operations domain_inode_ops = {
	lookup: domain_lookup,
};

static struct file_operations domain_file_ops = {
	read: generic_read_dir,
	readdir: domain_readdir,
};

static inline struct inode *mk_domain_inode(struct super_block *s,
	struct domain_inode_data *dd)
{
	struct inode *i = iget(s, allocate_inum(s));
	if(!i)
		goto fail_noinode;
	i->i_op = &domain_inode_ops;
	i->i_fop = &domain_file_ops;
	i->i_mode = S_IFDIR|0555;
	i->i_nlink = 2 + ARRAY_LEN(rrtype_names);
	DOMAIN_INODE_DATA(i) = dd;
	return i;
fail_noinode:
	kfree(dd);
	return ERR_PTR(-ENOMEM);
}

/* class directories */

static struct dentry *class_lookup(struct inode *parenti, struct dentry *childd)
{
	struct qstr const *name = &childd->d_name;
	int dnlen;
	struct domain_inode_data *dd;
	struct inode *di;
	if(name->len == 1 && *name->name == '\\') {
		static struct qstr const rootname = { name: ".", len: 1 };
		name = &rootname;
	}
	dnlen = domainname_text_to_bin(NULL, 0, name->name, name->len);
	if(dnlen == -1) {
		d_add(childd, NULL);
		return NULL;
	}
	dd = kmalloc(offsetof(struct domain_inode_data, domain_name)+dnlen,
			GFP_KERNEL);
	if(!dd)
		return ERR_PTR(-ENOMEM);
	dd->class = CLASS_INODE_DATA(parenti);
	domainname_text_to_bin(dd->domain_name, dnlen, name->name, name->len);
	di = mk_domain_inode(parenti->i_sb, dd);
	if(IS_ERR(di))
		return ERR_PTR(PTR_ERR(di));
	d_add(childd, di);
	return NULL;
}

static struct inode_operations class_inode_ops = {
	lookup: class_lookup,
};

static struct file_operations class_file_ops = {
	read: generic_read_dir,
};

static inline struct inode *mk_class_inode(struct super_block *s,
	domainclass_t class_number)
{
	struct class_inode_data *cd;
	struct inode *i;
	cd = kmalloc(sizeof(*cd), GFP_KERNEL);
	if(!cd)
		goto fail_nodata;
	cd->class_number = class_number;
	i = iget(s, allocate_inum(s));
	if(!i)
		goto fail_noinode;
	i->i_op = &class_inode_ops;
	i->i_fop = &class_file_ops;
	i->i_mode = S_IFDIR|0111;
	CLASS_INODE_DATA(i) = cd;
	return i;
fail_noinode:
	kfree(cd);
fail_nodata:
	return ERR_PTR(-ENOMEM);
}

/* root directory */

static struct dentry *root_lookup(struct inode *parenti, struct dentry *childd)
{
	int a, b;
	if(!namencmp(childd->d_name.name, "class", 5)) {
		struct inode *i;
		u16 n;
		if(scan_u16(childd->d_name.name + 5, &n))
			goto noent;
		i = mk_class_inode(parenti->i_sb, n);
		if(IS_ERR(i))
			return ERR_PTR(PTR_ERR(i));
		d_add(childd, i);
		return NULL;
	}
	for(a = 0, b = ARRAY_LEN(class_names); a != b; ) {
		int m = (a+b)/2;
		int cmp = namecmp(childd->d_name.name, class_names[m].name);
		if(!cmp) {
			/* TODO: do case folding more neatly -- get right inode number */
			struct inode *i = mk_class_inode(parenti->i_sb,
						class_names[m].class_number);
			if(IS_ERR(i))
				return ERR_PTR(PTR_ERR(i));
			d_add(childd, i);
			return NULL;
		}
		if(cmp < 0)
			b = m;
		else
			a = m+1;
	}
noent:
	d_add(childd, NULL);
	return NULL;
}

static int root_readdir(struct file *filp, void *dirent, filldir_t filldir)
{
	loff_t pos;
	for(pos = filp->f_pos; pos < 2+ARRAY_LEN(class_names); pos++) {
		if(pos == 0 || pos == 1) {
			if(filldir(dirent, "..", pos+1, pos, 1, DT_DIR))
				break;
		} else {
			ino_t inum;
			struct class_name const *cnp = &class_names[pos-2];
			char const *name = cnp->name;
			int namelen = strlen(name);
			struct dentry *de =
				lookup_one_len(name, filp->f_dentry, namelen);
			if(IS_ERR(de))
				return PTR_ERR(de);
			inum = de->d_inode->i_ino;
			dput(de);
			if(filldir(dirent, name, namelen, pos, inum, DT_DIR))
				break;
		}
	}
	filp->f_pos = pos;
	return 0;
}

static struct inode_operations root_inode_ops = {
	lookup: root_lookup,
};

static struct file_operations root_file_ops = {
	read: generic_read_dir,
	readdir: root_readdir,
};

static inline struct inode *mk_root_inode(struct super_block *s)
{
	struct inode *i = iget(s, 1);
	if(!i)
		return NULL;
	i->i_op = &root_inode_ops;
	i->i_fop = &root_file_ops;
	i->i_mode = S_IFDIR|0555;
	i->i_nlink = 2 + ARRAY_LEN(class_names);
	return i;
}

/* filesystem setup */

static void my_read_inode(struct inode *i)
{
	i->i_uid = 0;
	i->i_gid = 0;
	i->i_nlink = 1;
	i->i_size = 0;
	i->i_mtime = i->i_atime = i->i_ctime = CURRENT_TIME;
	i->i_blocks = i->i_blksize = 0;
}

static void my_put_inode(struct inode *i)
{
	if(i->i_op == &class_inode_ops) {
		struct class_inode_data *cd = CLASS_INODE_DATA(i);
		kfree(cd);
	} else if(i->i_op == &domain_inode_ops) {
		struct domain_inode_data *dd = DOMAIN_INODE_DATA(i);
		kfree(dd);
	} else if(i->i_op == &rrset_inode_ops) {
		struct rrset_inode_data *rd = RRSET_INODE_DATA(i);
		kfree(rd);
	}
}

static int my_statfs(struct super_block *sb, struct statfs *buf)
{
	buf->f_type = 0;
	buf->f_bsize = 1;
	buf->f_bfree = 0;
	buf->f_bavail = 0;
	buf->f_ffree = 0;
	buf->f_namelen = 0;
	return 0;
}

static void my_put_super(struct super_block *s)
{
	kfree(MY_SUPER_DATA(s));
}

static struct super_operations my_super_ops = {
	put_super: my_put_super,
	statfs: my_statfs,
	read_inode: my_read_inode,
	put_inode: my_put_inode,
};

static struct super_block *my_read_super(struct super_block *s, void *data,
	int silent)
{
	struct inode *root_inode;
	s->s_blocksize = 1;
	s->s_blocksize_bits = 0;
	s->s_op = &my_super_ops;
	s->s_magic = 0;
	MY_SUPER_DATA(s) = kmalloc(sizeof(struct my_super_data), GFP_KERNEL);
	if(!MY_SUPER_DATA(s))
		goto fail_nomydata;
	spin_lock_init(&MY_SUPER_DATA(s)->nextinode_lock);
	MY_SUPER_DATA(s)->nextinode = 2;
	root_inode = mk_root_inode(s);
	if(!root_inode)
		goto fail_norootinode;
	s->s_root = d_alloc_root(root_inode);
	if(!s->s_root)
		goto fail_norootdentry;
	return s;
fail_norootdentry:
	iput(root_inode);
fail_norootinode:
	kfree(MY_SUPER_DATA(s));
fail_nomydata:
	return NULL;
}

static DECLARE_FSTYPE(my_fs_type, "resolvfs", my_read_super, 0);

/* module information */

static int __init my_init(void)
{
	return register_filesystem(&my_fs_type);
}

static void __exit my_exit(void)
{
	unregister_filesystem(&my_fs_type);
}

module_init(my_init)
module_exit(my_exit)
EXPORT_NO_SYMBOLS;

#ifdef MODULE_AUTHOR
MODULE_AUTHOR("Andrew Main (Zefram) <zefram@fysh.org>");
#endif /* MODULE_AUTHOR */
#ifdef MODULE_LICENSE
MODULE_LICENSE("GPL");
#endif /* MODULE_LICENSE */
#ifdef MODULE_DESCRIPTION
MODULE_DESCRIPTION("domain name resolution virtual filesystem");
#endif /* MODULE_DESCRIPTION */
