/*
 * purestub.c -- pure stub DNS resolver for the Linux kernel
 * Copyright (C) 2001  Andrew Main
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <linux/module.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/stddef.h>
#include <linux/fs.h>
#include <linux/net.h>
#include <linux/in.h>
#include <net/sock.h>
#include <asm/types.h>
#include <asm/errno.h>
#include <asm/unaligned.h>
#include <asm/uaccess.h>
#include <asm/byteorder.h>
#include "resolv.h"
#include "resolv_numbers.h"

/*
 * overall TODO:
 *
 * + merge simultaneous identical queries
 *
 * + use UDP for initial queries
 *
 * + use a persistent TCP connection
 *
 * + timeout queries
 *
 * + be configurable re nameserver location; optionally use IPv6
 *
 * + recheck error handling code
 *
 * + generally clean up code
 *
 * + later, add OPTIONAL caching
 */

/* non-end-user data RRtypes */

#define T_NS		2
#define T_CNAME		5
#define T_SOA		6
#define T_SIG		24
#define T_KEY		25
#define T_NXT		30
#define T_A6		38
#define T_DNAME		39

/*
 * domain names in packets
 *
 * domainname_comp_check_length() checks the syntax of a compressed domain
 * name, including checking that it doesn't have a loop of compression
 * pointers.  It returns -1 if it finds bad syntax, or otherwise the
 * length of the first section, i.e., the amount that must be skipped
 * over to find the next thing in the packet.
 *
 * domainname_comp_bin_cmp() compares a compressed domain name with an
 * uncompressed binary domain name.  The ordering is consistent, but
 * not particularly meaningful.  Both domain names must be well-formed.
 *
 * domainname_comp_to_bin() uncompresses a domain name.  The input
 * must be well-formed.  The output buffer is pointed to by `buf',
 * and is of length `buflen'.  If there is enough space, the binary
 * form of the domain name is written into the start of the buffer,
 * the rest of the buffer is untouched, and the length of the output
 * is returned.  If there is insufficient space, the contents of the
 * buffer is unspecified, and the length of the output that would have
 * been produced (>buflen) is returned.
 *
 * TODO: domainname_bin_to_comp()
 */

static u32 _comp_label_class_and_length(u8 const *packet, size_t packetlen,
	ssize_t p)
{
	u8 o;
	if(p < 0 || (size_t)p >= packetlen)
		return (u32)-1;
	o = packet[p];
	if(!o)
		return (0<<16) + 1;
	else if(o < 64) {
		p += 1+o;
		if(p < 0 || (size_t)p > packetlen)
			return (u32)-1;
		return (1<<16) + 1+o;
	} else if((o & 0xc0) == 0xc0) {
		p += 2;
		if(p < 0 || (size_t)p > packetlen)
			return (u32)-1;
		return (2<<16) + 2;
	} else
		return (u32)-1;
}

static u32 _bin_label_class_and_length(u8 const *bin)
{
	u32 lc = _comp_label_class_and_length(bin, (size_t)-1, 0);
	if((lc>>16) == 2)
		lc = (u32)-1;
	return lc;
}

static ssize_t _follow_ptr(u8 const *packet, size_t packetlen, ssize_t p)
{
	return ntohs(get_unaligned((u16 *)(packet+p))) & 0x3fff;
}

static inline u8 _label_lower(u8 c)
{
	if(c >= 'A' && c <= 'Z')
		c += 'a'-'A';
	return c;
}

static int _label_cmp(u8 const *a, u8 const *b)
{
	u8 la = *a, lb = *b;
	if(la != lb)
		return (int)la-(int)lb;
	for(; la--; ) {
		u8 ca = _label_lower(*++a);
		u8 cb = _label_lower(*++b);
		if(ca != cb)
			return (int)ca-(int)cb;
	}
	return 0;
}

static ssize_t _section_end(u8 const *packet, size_t packetlen, ssize_t p,
	u32 *clp)
{
	while(1) {
		u32 cl = _comp_label_class_and_length(packet, packetlen, p);
		if((cl>>16) != 1) {
			*clp = cl;
			return p;
		}
		p += cl & 0xffff;
	}
}

static ssize_t _next_section(u8 const *packet, size_t packetlen, ssize_t p)
{
	u32 cl;
	if(p < 0)
		return p;
	p = _section_end(packet, packetlen, p, &cl);
	if((cl>>16) == 0)
		return -2;
	else if((cl>>16) == 2)
		return _follow_ptr(packet, packetlen, p);
	else
		return -1;
}

static ssize_t domainname_comp_check_length(u8 const *packet, size_t packetlen,
	ssize_t comp)
{
	ssize_t ret, fast, slow;
	u32 cl;
	fast = _section_end(packet, packetlen, comp, &cl);
	if(cl == (u32)-1)
		return -1;
	ret = fast+(cl&0xffff) - comp;
	if((cl>>16) == 0)
		return ret;
	slow = fast = _follow_ptr(packet, packetlen, fast);
	while(1) {
		fast = _next_section(packet, packetlen, _next_section(packet, packetlen, fast));
		if(fast == -2)
			return ret;
		if(fast == -1 || fast == slow)
			return -1;
		slow = _next_section(packet, packetlen, slow);
	}
}

static ssize_t domainname_bin_length(u8 const *dn)
{
	ssize_t ret = 0;
	while(1) {
		u32 cl = _bin_label_class_and_length(dn);
		if(cl == (u32)-1)
			return -1;
		ret += cl & 0xffff;
		dn += cl & 0xffff;
		if(!(cl>>16))
			return ret;
	}
}

static ssize_t domainname_bin_check_length(u8 const *dn, size_t maxlen)
{
	ssize_t ret = 0;
	while(1) {
		u32 cl = _comp_label_class_and_length(dn, maxlen, ret);
		if(cl == (u32)-1 || (cl>>16) == 2)
			return -1;
		ret += cl & 0xffff;
		if(!(cl>>16))
			return ret;
	}
}

static int domainname_comp_bin_cmp(u8 const *packet, size_t packetlen,
	ssize_t comp, u8 const *bin)
{
	while(1) {
		int cmp;
		u32 clc;
		u32 clb = _bin_label_class_and_length(bin);
		while(1) {
			clc = _comp_label_class_and_length(packet, packetlen,
				comp);
			if((clc>>16) != 2)
				break;
			comp = _follow_ptr(packet, packetlen, comp);
		}
		if((clc>>16) != (clb>>16))
			return (int)(clc>>16) - (int)(clb>>16);
		if(!(clc>>16))
			return 0;
		cmp = _label_cmp(packet+comp, bin);
		if(cmp)
			return cmp;
		comp += clc&0xffff;
		bin += clb&0xffff;
	}
}

static ssize_t domainname_comp_to_bin(u8 *buf, size_t buflen,
	u8 const *packet, size_t packetlen, ssize_t comp)
{
	ssize_t olen = 0;
	while(1) {
		u32 cl;
		size_t llen;
		while(1) {
			cl = _comp_label_class_and_length(packet, packetlen,
				comp);
			if((cl>>16) != 2)
				break;
			comp = _follow_ptr(packet, packetlen, comp);
		}
		llen = cl & 0xffff;
		if(olen + llen <= buflen)
			memcpy(buf+olen, packet+comp, llen);
		olen += llen;
		if(!(cl>>16))
			return olen;
		comp += llen;
	}
}

/*
 * socket I/O
 */

struct socket *open_tcp_conn(void)
{
	struct socket *sock;
	struct sockaddr_in nsaddr;
	int result = sock_create(PF_INET, SOCK_STREAM, IPPROTO_TCP, &sock);
	if(result < 0)
		goto err_nosocket;
	nsaddr.sin_family = AF_INET;
	nsaddr.sin_port = htons(53);
	nsaddr.sin_addr.s_addr = htonl(INADDR_LOOPBACK);
	result = sock->ops->connect(sock, (struct sockaddr *)&nsaddr,
						sizeof(nsaddr), O_RDWR);
	if(result < 0)
		goto err_noconnect;
	return sock;
err_noconnect:
	sock_release(sock);
err_nosocket:
	return ERR_PTR(result);
}

static ssize_t sock_retrying_write(struct socket *sock,
	void const *data, size_t length)
{
	mm_segment_t oldfs;
	struct msghdr msgh;
	struct iovec iovec;
	ssize_t result;
	msgh.msg_name = NULL;
	msgh.msg_namelen = 0;
	msgh.msg_iov = &iovec;
	msgh.msg_iovlen = 1;
	msgh.msg_control = NULL;
	msgh.msg_controllen = 0;
	msgh.msg_flags = MSG_NOSIGNAL;
	iovec.iov_base = (void *)data;
	iovec.iov_len = length;
	oldfs = get_fs(); set_fs(KERNEL_DS);
	for(;;) {
		result = sock_sendmsg(sock, &msgh, iovec.iov_len);
		if(result < 0 || result == iovec.iov_len)
			break;
		iovec.iov_base = ((u8 *)iovec.iov_base) + result;
		iovec.iov_len -= result;
	}
	set_fs(oldfs);
	return result < 0 ? result : 0;
}

static ssize_t sock_retrying_read(struct socket *sock, void *buf, size_t buflen)
{
	mm_segment_t oldfs;
	struct msghdr msgh;
	struct iovec iovec;
	int result;
	msgh.msg_name = NULL;
	msgh.msg_namelen = 0;
	msgh.msg_iov = &iovec;
	msgh.msg_iovlen = 1;
	msgh.msg_control = NULL;
	msgh.msg_controllen = 0;
	msgh.msg_flags = 0;
	iovec.iov_base = buf;
	iovec.iov_len = buflen;
	oldfs = get_fs(); set_fs(KERNEL_DS);
	result = sock_recvmsg(sock, &msgh, iovec.iov_len, MSG_WAITALL);
	set_fs(oldfs);
	return result < 0 ? result : result == buflen ? 0 : -EPROTO;
}

/*
 */

static int build_query_packet(domainname_ptr_t domainname,
	domainclass_t domainclass, rrtype_t rrtype,
	u8 **msg_p, size_t *msglen_p)
{
	u8 *query;
	size_t dnl = domainname_bin_length(domainname);
	size_t msglen = 14+dnl+4;
	if(msglen-2 > (size_t)((u16)-1))
		return -EPROTO;
	query = kmalloc(msglen, GFP_KERNEL);
	if(!query)
		return -ENOMEM;
	*(u16 *)query = htons(msglen-2);
	*(u16 *)(query+2) = htons(0);                   /* ID */
	*(u16 *)(query+4) = htons(0x0100);              /* flags */
	*(u16 *)(query+6) = htons(1);                   /* QDcount */
	*(u16 *)(query+8) = htons(0);                   /* ANcount */
	*(u16 *)(query+10) = htons(0);                  /* NScount */
	*(u16 *)(query+12) = htons(0);                  /* ARcount */
	memcpy(query+14, domainname, dnl);              /* Qname */
	put_unaligned(htons(rrtype), (u16 *)(query+14+dnl));
							/* Qtype */
	put_unaligned(htons(domainclass), (u16 *)(query+14+dnl+2));
							/* Qclass */
	*msg_p = query;
	*msglen_p = msglen;
	return 0;
}

static int read_reply(struct socket *sock, u8 **reply_p, size_t *replylen_p)
{
	u16 replylen;
	u8 *reply;
	int result;
	result = sock_retrying_read(sock, &replylen, 2);
	replylen = ntohs(replylen);
	if(result)
		goto err;
	reply = kmalloc(replylen, GFP_KERNEL);
	if(!reply)
		goto err_enomem;
	result = sock_retrying_read(sock, reply, replylen);
	if(result)
		goto err_free;
	*reply_p = reply;
	*replylen_p = replylen;
	return 0;
err_free:
	kfree(reply);
err:
	return result;
err_enomem:
	return -ENOMEM;
}

static int dnspkt_is_valid(u8 *packet, size_t packetlen)
{
	size_t p;
	u16 qdcount;
	u32 rrcount;
	if(packetlen < 12)
		return 0;
	qdcount = ntohs(*(u16 *)(packet+4));
	rrcount = (u32)ntohs(*(u16 *)(packet+6)) +
		(u32)ntohs(*(u16 *)(packet+8)) +
		(u32)ntohs(*(u16 *)(packet+10));
	p = 12;
	while(qdcount--) {
		ssize_t nlen =
			domainname_comp_check_length(packet, packetlen, p);
		if(nlen < 0)
			return 0;
		p += nlen;
		if(packetlen-p < 4)
			return 0;
		p += 4;
	}
	while(rrcount--) {
		ssize_t nlen =
			domainname_comp_check_length(packet, packetlen, p);
		if(nlen < 0)
			return 0;
		p += nlen;
		if(packetlen-p < 10)
			return 0;
		nlen = ntohs(get_unaligned((u16 *)(packet+p+8)));
		p += 10;
		if(packetlen-p < nlen)
			return 0;
		p += nlen;
	}
	return p == packetlen;
}

static int dnspkt_answers_query(u8 const *msg, size_t msglen,
	domainname_ptr_t domainname, domainclass_t domainclass, rrtype_t rrtype)
{
	size_t pos;
	if(ntohs(*(u16 *)(msg+4)) != 1 ||
			(ntohs(*(u16 *)(msg+2)) & 0xfa80) != 0x8080 ||
			domainname_comp_bin_cmp(msg, msglen, 12, domainname))
		return 0;
	pos = 12 + domainname_comp_check_length(msg, msglen, 12);
	if(ntohs(get_unaligned((u16 *)(msg+pos))) != rrtype ||
			ntohs(get_unaligned((u16 *)(msg+pos+2))) != domainclass)
		return 0;
	return 1;
}

static int do_query(domainname_ptr_t domainname,
	domainclass_t domainclass, rrtype_t rrtype,
	u8 **reply_p, size_t *replylen_p)
{
	struct socket *qsock;
	u8 *msg;
	size_t msglen;
	int result;
	qsock = open_tcp_conn();
	if(IS_ERR(qsock)) {
		result = PTR_ERR(qsock);
		goto err_nosocket;
	}
	result = build_query_packet(domainname, domainclass, rrtype,
			&msg, &msglen);
	if(result < 0)
		goto err_noquery;
	result = sock_retrying_write(qsock, msg, msglen);
	kfree(msg);
	if(result < 0)
		goto err_noquery;
	result = qsock->ops->shutdown(qsock, 1);
	if(result < 0)
		goto err_noquery;
	result = read_reply(qsock, &msg, &msglen);
	sock_release(qsock);
	if(result < 0)
		goto err_noreply;
	if(!dnspkt_is_valid(msg, msglen) ||
			!dnspkt_answers_query(msg, msglen,
				domainname, domainclass, rrtype))
		goto err_badreply;
	*reply_p = msg;
	*replylen_p = msglen;
	return 0;
err_badreply:
	kfree(msg);
	result = -EPROTO;
err_noreply:
	return result;
err_noquery:
	sock_release(qsock);
err_nosocket:
	return result;
}

static ssize_t dnspkt_extract_rr(u8 *buf, size_t buflen,
	u8 *packet, size_t packetlen, size_t rdpos, size_t rdlen,
	rrtype_t rrtype)
{
	/*
	 * In extracting an RR from a packet, we need to expand embedded
	 * domain names, because compressed domain names won't make
	 * any sense outside the context of the packet.  Domain name
	 * expansion is mandatory for the well-known RRtypes of RFC
	 * 1035, but advisable for all known types, since some DNS
	 * implementations compress domain names in other RRs.
	 *
	 * All known types have all their embedded domain names
	 * consecutive.  The domain names are at different offsets in
	 * different types, in some cases at variable offsets, and may
	 * have trailing non-domain data.  In some cases domain names
	 * are optional.  We generalise all of this to treating the
	 * Rdata as some number of octets of plain data, followed by some
	 * number of domain names, followed by arbitrary additional data.
	 * The big switch statement below determines the parameters:
	 * offset and number of domain names.  The code after the switch
	 * does the extraction and expansion, gracefully handling Rdata
	 * that doesn't have everything the parameters suggest.
	 */
	size_t domainoffset, ndomains, olen;
	switch(rrtype) {
		case T_NS:
		case T_MD:
		case T_MF:
		case T_CNAME:
		case T_MB:
		case T_MG:
		case T_MR:
		case T_PTR:
		case T_NXT:
		case T_DNAME:
			domainoffset = 0;
			ndomains = 1;
			break;
		case T_SOA:
		case T_MINFO:
		case T_RP:
			domainoffset = 0;
			ndomains = 2;
			break;
		case T_MX:
		case T_AFSDB:
		case T_RT:
			domainoffset = 2;
			ndomains = 1;
			break;
		case T_SIG:
			domainoffset = 18;
			ndomains = 1;
			break;
		case T_PX:
			domainoffset = 2;
			ndomains = 2;
			break;
		case T_SRV:
			domainoffset = 6;
			ndomains = 1;
			break;
		case T_NAPTR: {
			int n;
			domainoffset = 4;
			ndomains = 1;
			for(n = 3; n-- && rdlen <= domainoffset; )
				domainoffset += 1 + packet[rdpos+domainoffset];
		}
		case T_A6: {
			u8 plen;
			if(!rdlen || (plen = packet[rdpos]) > 128) {
				domainoffset = 0;
				ndomains = 0;
			} else {
				domainoffset = 1 + ((128-plen+7) >> 3);
				ndomains = 1;
			}
			break;
		}
		default:
			domainoffset = 0;
			ndomains = 0;
			break;
	}
	if(rdlen < domainoffset)
		domainoffset = rdlen;
	if(buflen >= domainoffset)
		memcpy(buf, packet+rdpos, domainoffset);
	olen = domainoffset;
	rdpos += domainoffset;
	rdlen -= domainoffset;
	while(ndomains--) {
		size_t clen = (size_t)domainname_comp_check_length(packet,
						packetlen, rdpos);
		if(rdlen < clen)
			break;
		olen += domainname_comp_to_bin(buflen >= olen ? buf+olen : NULL,
				buflen >= olen ? buflen-olen : 0,
				packet, packetlen, rdpos);
		rdpos += clen;
		rdlen -= clen;
	}
	if(buflen >= olen+rdlen)
		memcpy(buf+olen, packet+rdpos, rdlen);
	olen += rdlen;
	return olen;
}

static struct rrset *dnspkt_extract_answers(u8 *packet, size_t packetlen,
	domainname_ptr_t domainname, domainclass_t domainclass, rrtype_t rrtype)
{
	u16 qdcount = ntohs(*(u16 *)(packet+4));
	u16 ancount = ntohs(*(u16 *)(packet+6));
	size_t p = 12;
	struct rrset *buf = kmalloc(sizeof(struct rrset), GFP_KERNEL);
	if(!buf)
		return ERR_PTR(-ENOMEM);
	buf->size = offsetof(struct rrset, rdata);
	buf->rrcount = 0;
	while(qdcount--)
		p += domainname_comp_check_length(packet, packetlen, p) + 4;
	while(ancount--) {
		size_t dp = p, rdlen;
		p += domainname_comp_check_length(packet, packetlen, p);
		rdlen = ntohs(get_unaligned((u16 *)(packet+p+8)));
		if(ntohs(get_unaligned((u16 *)(packet+p+2))) == domainclass &&
			ntohs(get_unaligned((u16 *)(packet+p+0))) == rrtype &&
				!domainname_comp_bin_cmp(packet, packetlen, dp,
					domainname)) {
			struct rrset *newbuf;
			struct rdata *rd;
			ssize_t olen = dnspkt_extract_rr(NULL, 0,
					packet, packetlen, p+10, rdlen, rrtype);
			if(olen > 0xffff)
				olen = -EPROTO;
			if(olen < 0) {
				kfree(buf);
				return ERR_PTR(olen);
			}
			newbuf = kmalloc(buf->size+sizeof(struct rdata)+olen,
						GFP_KERNEL);
			if(!newbuf) {
				kfree(buf);
				return ERR_PTR(-ENOMEM);
			}
			memcpy(newbuf, buf, buf->size);
			kfree(buf);
			buf = newbuf;
			rd = (struct rdata *)(((u8 *)buf)+buf->size);
			put_unaligned(olen, &rd->length);
			dnspkt_extract_rr(rd->data, olen,
					packet, packetlen, p+10, rdlen, rrtype);
			buf->size += sizeof(struct rdata) + olen;
			buf->rrcount++;
		}
		p += 10 + rdlen;
	}
	return buf;
}

static struct rrset *external_query(domainname_ptr_t domainname,
	domainclass_t domainclass, rrtype_t rrtype)
{
	u8 *reply;
	size_t replylen;
	int result;
	struct rrset *rrs;
	void *tofree_domainname = NULL;
	int cnames_permitted = 16;
	int retry_on_nodata;
make_query:
	result = do_query(domainname, domainclass, rrtype, &reply, &replylen);
	if(result < 0)
		return ERR_PTR(result);
	result = -ENOENT;
	if((ntohs(*(u16 *)(reply+2)) & 0x000f) == 3)
		goto err_badreply;   /* NXDOMAIN */
	result = -EPROTO;
	if((ntohs(*(u16 *)(reply+2)) & 0x000f) != 0)
		goto err_badreply;
	retry_on_nodata = 0;
check_answers:
	rrs = dnspkt_extract_answers(reply, replylen,
		domainname, domainclass, rrtype);
	if(IS_ERR(rrs))
		goto err_norrs;
	if(!rrs->rrcount) {
		kfree(rrs);
		rrs = dnspkt_extract_answers(reply, replylen,
			domainname, domainclass, T_CNAME);
		if(IS_ERR(rrs))
			goto err_norrs;
		if(rrs->rrcount) {
			result = -EPROTO;
			if(!cnames_permitted-- ||
				domainname_bin_check_length(rrs->rdata[0].data,
						rrs->rdata[0].length) !=
					rrs->rdata[0].length)
				goto err_badreply;
			kfree(tofree_domainname);
			tofree_domainname = rrs;
			domainname = rrs->rdata[0].data;
			retry_on_nodata = 1;
			goto check_answers;
		} else {
			/* NODATA, or ran out of space for CNAMEs */
			if(retry_on_nodata) {
				kfree(rrs);
				kfree(reply);
				goto make_query;
			}
		}
	}
	kfree(tofree_domainname);
	kfree(reply);
	return rrs;
err_norrs:
	result = PTR_ERR(rrs);
err_badreply:
	kfree(tofree_domainname);
	kfree(reply);
	return ERR_PTR(result);
}

/*
 * class and RRtype classification
 *
 * Class and RRtype numbers fall into several categories:
 *
 * * meta-types: have special meaning in the DNS protocol, not used
 *   for data
 * * non-end-user data types: cacheable data, but metadata having meaning
 *   only within participants to the DNS protocol, not to end users
 * * end-user data types: normal data that is meaningful to end users
 * * unallocated types reserved for meta-types
 * * unallocated types reserved for data types
 * * unallocated types with no reservation
 *
 * We want to allow lookups only on end-user data.  This means we must
 * reject requests for meta-types, unknown types reserved for meta-types,
 * and known non-end-user data types.  The standard practice with other
 * unknown types is to treat them as end-user data types until they
 * become known.
 */

static int is_enduser_data_class(domainclass_t domainclass)
{
	return domainclass < 128 || domainclass >= 256;
}

static int is_enduser_data_type(rrtype_t rrtype)
{
	if(rrtype < 32)
		return 1 & (0xbcffff9a >> rrtype);
	if(rrtype < 64)
		return 1 & (0xfffffd3f >> rrtype);
	return rrtype < 128 || rrtype >= 256;
}

/*
 * main resolver entry points
 */

int resolv_userrset(domainname_ptr_t domainname,
	domainclass_t domainclass, rrtype_t rrtype,
	int (*cb)(void *cb_magic, struct rrset const *rrset), void *cb_magic)
{
	struct rrset *rrs;
	int result;
	if(!is_enduser_data_class(domainclass) || !is_enduser_data_type(rrtype))
		return -EINVAL;
	rrs = external_query(domainname, domainclass, rrtype);
	if(IS_ERR(rrs))
		return PTR_ERR(rrs);
	result = cb(cb_magic, rrs);
	kfree(rrs);
	return result;
}

/*
 * resolv_getrrset() wrapper
 *
 * TODO: do more efficiently to avoid copying
 */

static int getrrset_callback(void *cb_magic, struct rrset const *rrset)
{
	size_t sz = rrset->size;
	void *output_rrset = kmalloc(sz, GFP_KERNEL);
	if(!output_rrset)
		return -ENOMEM;
	memcpy(output_rrset, rrset, sz);
	*((struct rrset **)cb_magic) = (struct rrset *)output_rrset;
	return 0;
}

struct rrset *resolv_getrrset(domainname_ptr_t domainname,
	domainclass_t domainclass, rrtype_t rrtype)
{
	struct rrset *output_rrset;
	int err = resolv_userrset(domainname, domainclass, rrtype,
					getrrset_callback, &output_rrset);
	return err ? ERR_PTR(err) : output_rrset;
}

/*
 * module stuff
 */

static int __init my_init(void)
{
	return 0;
}

static void __exit my_exit(void)
{
}

module_init(my_init)
module_exit(my_exit)
EXPORT_SYMBOL(resolv_getrrset);
EXPORT_SYMBOL(resolv_userrset);

#ifdef MODULE_AUTHOR
MODULE_AUTHOR("Andrew Main (Zefram) <zefram@fysh.org>");
#endif /* MODULE_AUTHOR */
#ifdef MODULE_LICENSE
MODULE_LICENSE("GPL");
#endif /* MODULE_LICENSE */
#ifdef MODULE_DESCRIPTION
MODULE_DESCRIPTION("Stub DNS resolver for the kernel");
#endif /* MODULE_DESCRIPTION */
